/* audio_win32.c */

/* created from */
/* audio player */
/* low level waveform audio services */
/* create Thu Feb 11 23:07:38 JST 1999 */
/* Copyright (C) 1999, by linear a. All rights reserved. */

/* assum: sizeof(short) == 16bits */

#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <wtypes.h>
#include <windowsx.h>
#include <mmsystem.h>
#include <conio.h>
#include "all.h"

extern unsigned long  AIFF_sampleRate;
int l = 0, r = 0;

/* some debugging aid */
#define dprintf(a) printf("[DBG] "); printf a
#define MMFUNC(func, param) \
	if ((func param) != MMSYSERR_NOERROR) { \
		fprintf(stderr, "failed: " #func " (" __FILE__ ":%d)\n", __LINE__); \
		exit(1); \
	}

#ifndef TRUE
#define TRUE 1
#define FALSE 0
#endif

#define DATA_TOP 0x28
#define WAVE_HEADER_SIZE 44

typedef struct {
    unsigned int dev_list;
    unsigned int file_info;
    unsigned int gain;
    double gain_fact;
    FILE *playfp;
    unsigned char *playbuf[8];
    int buf_sz;			/* buffer size */
} global_t;

typedef struct {
    unsigned int start, stop;
} window_t;

global_t glo;

#define BLOCK 4
#define NBUF 8			/* must be power of 2 */
/* 20:8 */

PCMWAVEFORMAT pcm;
HANDLE event;
HANDLE inev[2];

void die(unsigned char *s)
{
    fprintf(stderr, "%s", s);
    exit(1);
}

/* get little endian data */
unsigned int get_leb(unsigned char **p)
{
    unsigned int i;
    i = (*p)[0];
    (*p)++;
    return i;
}

unsigned int get_lew(unsigned char **p)
{
    unsigned int i;
    i = ((*p)[1] << 8) + (*p)[0];
    (*p) += 2;
    return i;
}

unsigned int get_lel(unsigned char **p)
{
    unsigned int i;
    i = (((((*p)[3] << 8) + (*p)[2]) << 8) + (*p)[1] << 8) + (*p)[0];
    (*p) += 4;
    return i;
}

/* put little endian data */
void put_leb(unsigned char **p, unsigned int i)
{
    unsigned char c = i;
    (*p)++[0] = c;
}

void put_lew(unsigned char **p, unsigned int i)
{
    unsigned char c = i;
    (*p)++[0] = c;
    c = i >> 8;
    (*p)++[0] = c;
}

void put_lel(unsigned char **p, unsigned int i)
{
    unsigned char c = i;
    (*p)++[0] = c;
    i >>= 8;
    c = i;
    (*p)++[0] = c;
    i >>= 8;
    c = i;
    (*p)++[0] = c;
    i >>= 8;
    c = i;
    (*p)++[0] = c;
}


#define my_gmalloc(a) gmalloc((a), GMEM_FIXED)

unsigned char *gmalloc(int size, int attr)
{
    unsigned char *buf = GlobalAllocPtr(attr, size);
    if (!buf) {
	die("no sufficiet global memory\n");
    }
    return buf;
}

/* We make some small functions, for exercise */

/* Windows wav format handler */
/* all data is little endian
   0 c4		"RIFF"
   4 L		file size
   8 c4 	"WAVE"
   c c4		"fmt "
   14 ?
   16 W		channel
   18 L		sample/sec
   1c L		bytes/sec
   20 W		block align
   22 W		bits/sample
   24 c4	"data"
   28 L		size
*/

PCMWAVEFORMAT pcm;
int n, i, last;
int wait[NBUF];
WAVEHDR h[NBUF];

void open_outdev(PCMWAVEFORMAT *pcm, int odev);

void set_wav_header(int odev)
{
  int i;
    unsigned char header[DATA_TOP];
    unsigned char *p = &header[0x16];

    pcm.wf.nChannels = 2;
    pcm.wf.nSamplesPerSec = AIFF_sampleRate;
    pcm.wf.nAvgBytesPerSec = pcm.wf.nSamplesPerSec * 4;
    pcm.wf.nBlockAlign = 4;
    pcm.wBitsPerSample = 16;

    glo.buf_sz = 4096 * BLOCK;
    open_outdev(&pcm, odev);
    for (i = 0; i < NBUF; i++) {
      wait[i] = FALSE;
      glo.playbuf[i] = my_gmalloc(glo.buf_sz);
    }
    if (FCenter == 1) {
      l = 1;
      r = 2;
    } else {
      l = 0;
      r = 1;
    }
}


/* some global vars. */
int indevs;
int outdevs;
HWAVEIN wavin;
HWAVEOUT wavout;

int get_ndevs()
{
    int i = indevs = waveInGetNumDevs();
    int o = outdevs = waveOutGetNumDevs();
    dprintf(("in %d, out %d\n", i, o));
    return 0;
}

/* print dev caps info (cross section of WAVEINCAPSand WAVEOUTCAPS)*/
void print_dev_caps_title(void)
{
    if (glo.dev_list) {
	printf("I/O\tDID\tMID:PID\tVer.\tCh.\t\tname\n");
    }
}

void print_dev_caps(void *p, unsigned char *io, int did, int flag)
{
    WAVEOUTCAPS *in = p;
    if (glo.dev_list) {
	printf("%s%s\t%d\t%d:%d\t%d.%d\t%d\t\t%s\n",
	       io,
	       flag ? "*" : "",
	       did,
	       in->wMid, in->wPid, 
	       in->vDriverVersion >> 8, in->vDriverVersion & 255, 
	       in->wChannels,
	       in->szPname);
    }
}



void print_wav_header(PCMWAVEFORMAT *pcm)
{
  printf("sampling rate %d Hz %d bits %s (%d bytes/sec, align %d)\n", 
	 pcm->wf.nSamplesPerSec,
	 pcm->wBitsPerSample,
	 (pcm->wf.nChannels == 2) ? "stereo" : "mono",
	 pcm->wf.nAvgBytesPerSec,
	 pcm->wf.nBlockAlign
	 );
}

void open_outdev(PCMWAVEFORMAT *pcm, int odev)
{
    pcm->wf.wFormatTag = WAVE_FORMAT_PCM;
    print_wav_header(pcm);

    event = CreateEvent(NULL, TRUE, FALSE, NULL);
    if (waveOutOpen(&wavout, odev, (LPCWAVEFORMATEX) pcm, event, 0, CALLBACK_EVENT) != MMSYSERR_NOERROR) {
      fprintf(stderr, "invalid output device. use wave mapper.\n");
      MMFUNC(waveOutOpen, (&wavout, WAVE_MAPPER, (LPCWAVEFORMATEX) pcm, event, 0, CALLBACK_EVENT));
    }
    dprintf(("wavout %d\n", wavout));
}

int play_wav(WAVEHDR *h)
{
    int i;

    h->dwFlags = 0;
    waveOutPrepareHeader(wavout, h, sizeof(WAVEHDR));
    ResetEvent(event);
    i = waveOutWrite(wavout, h, sizeof(WAVEHDR));
    /*    dprintf(("return %d buf %p, played %d bytes\n", i, h->lpData, h->dwBufferLength));*/
    return 0;
}

static int count = -1;
#define last (count & (NBUF - 1))
#define next ((count + 1) & (NBUF - 1))
int play(unsigned char *p, int n)
{
  static int b = 0;

  /* wait buf */
  while (wait[next] && !(h[next].dwFlags & WHDR_DONE)) {
    WaitForSingleObject(event, INFINITE);
  }
  wait[next] = FALSE;
  /* blocking */
  memcpy(glo.playbuf[next] + n * b, p, n);
  b++;
  if (b < BLOCK) {
    return 0;
  }
  b = 0;
  /* prepare buf */
  h[next].lpData = glo.playbuf[next];
  h[next].dwBufferLength = n * BLOCK;
  play_wav(&h[next]);
  wait[next] = TRUE;
  count++;
  return 0;
}

void wait_last()
{
  int i;
  /* wait last buf */
  for (i = 0; i < BLOCK; i++) {
    if (wait[next + i]) {
      while (!(h[next + i].dwFlags & WHDR_DONE)) {
	WaitForSingleObject(event, INFINITE);
      }
      ResetEvent(event);
    }
  }
}

